/*
  Originally written by Naoto Kaneko
  Last updated: 3.Jul.2010, buggy-and-messy version
  Latest info: http://www37.atwiki.jp/pjkyt/
 */

package com.kyt.android.example;

import android.content.*;
import android.graphics.*;
import android.os.Vibrator;
import android.util.*;

import java.util.*;

import com.kyt.android.view.*;

/**
 * This example code demonstrates a few things
 * 
 * 1. Rendering of graphic objects that have coordinate in the ground plane (BubbleNodes).
 *    Their position in the view port depends on the location of the window.
 *    DrawNodeList is used to store nodes and to make the list of node to be rendered
 *    -> See drawBubbleNodes() and BubbleNode.java
 *    
 * 2. Rendering of graphic objects that have coordinate in the screen (CtrlNodes).
 *    They remain still at the same position of the view port.
 *    Regular array is used to store nodes since they are anyway drawn all the time
 *    -> See drawCtrlNodes(Canvas, Paint) and CtrlNode.java
 * 
 * 3. Moving window around with slide and flick gesture
 * 		-> This is the feature provided by ViewportedView automatically
 * 
 * 4. Zoom in and out -- i.e. actually shrinking and expanding window.
 *    -> See tapCtrlNode()
 * 
 * 5. Selection of a node with a tap
 *    A tap on the CtrlNodes triggers zoom.  And a top the BubbleNode toggles color.
 *    -> See onTouch(), tapCtrlNode(), and tapBubbleNode()
 * 
 * 6. Dragging of a node 
 *    A long press selects the node to drag
 *    -> See onTouch() and dragBubbleNode()
 *    
 * 7. Toggling the position indicator
 *    -> onDraw()
 */

public class VpvExampleView extends ViewportedView {
	private final int BG_COLOR= Color.BLACK;
	
	private static int GroundPlaneSize = 1000;
	private DrawWindow Window;
	
	private DrawNodeList BubbleNodes; // all BubbleNode
	private ArrayList<DrawNode> BubbleNodesToDraw; // BubbleNode that needs to be drawn with current window

	private CtrlNode[] CtrlNodes; // holds zoom in/out buttons
	
	private Vibrator Vibra;
	
	public VpvExampleView(Context context) {
		super(GroundPlaneSize, context);
		setBackgroundColor(BG_COLOR);
		
		Window = getDrawWindow();
		
		loadBubbleNodes();
		makeCtrlNodes();
		
		Vibra = (Vibrator)context.getSystemService(Context.VIBRATOR_SERVICE);
	}
		
	private void loadBubbleNodes() {
		BubbleNodes = new DrawNodeList(Window);

		// BubbleNode holds ground plane coordinate
		BubbleNodes.addNode(new BubbleNode("A", 20,30));
		BubbleNodes.addNode(new BubbleNode("B", 20,60));
		BubbleNodes.addNode(new BubbleNode("C", 50,120));
		BubbleNodes.addNode(new BubbleNode("D", 140,50));
		BubbleNodes.addNode(new BubbleNode("E", 200,330));
		BubbleNodes.addNode(new BubbleNode("F", 430,100));
		BubbleNodes.addNode(new BubbleNode("G", 500,500));
		BubbleNodes.addNode(new BubbleNode("H", 750,300));
		BubbleNodes.addNode(new BubbleNode("I", 900,800));
		BubbleNodes.addNode(new BubbleNode("J", 800,900));
	}
		
	private final int ZOOM_IN_INDEX = 0;
	private final int ZOOM_OUT_INDEX = 1;
	
	private void makeCtrlNodes() {
		// make control buttons
		// the coord of CtrlNode is that of the view port
		int screen_width = Window.getViewPortWidth();
		int screen_height = Window.getViewPortHeight();
		
		CtrlNodes = new CtrlNode[2];
		CtrlNodes[ZOOM_IN_INDEX] = new CtrlNode("+",0,0);
		CtrlNodes[ZOOM_IN_INDEX].setX(screen_width-4*CtrlNodes[ZOOM_IN_INDEX].getRadius());
		CtrlNodes[ZOOM_IN_INDEX].setY(screen_height-CtrlNodes[ZOOM_IN_INDEX].getRadius()-10);
		CtrlNodes[ZOOM_OUT_INDEX] = new CtrlNode("-",0,0);
		CtrlNodes[ZOOM_OUT_INDEX].setX(screen_width-(int)(1.5*CtrlNodes[ZOOM_IN_INDEX].getRadius()));
		CtrlNodes[ZOOM_OUT_INDEX].setY(screen_height-CtrlNodes[ZOOM_IN_INDEX].getRadius()-10);
	}
	
	// -- Draw related ------------------------------------------------------------
	
	private final int POSITION_BAR_COLOR = Color.GREEN;
	
	@Override
	protected void onDraw(Canvas canvas) {
		Paint p = new Paint();
		
		drawBubbleNodes(canvas, p);
		drawCtrlNodes(canvas, p);
		super.drawPositionIndicator(canvas, p, POSITION_BAR_COLOR);
	}
	
	private void drawBubbleNodes(Canvas canvas, Paint p) {
		BubbleNodesToDraw = BubbleNodes.getNodesToDraw();
		int list_length = BubbleNodesToDraw.size();
		
		for(int n=0; n!=list_length; n++) {
			BubbleNodesToDraw.get(n).draw(canvas, p, Window);
		}	
	}
	
	private void drawCtrlNodes(Canvas canvas, Paint p) {		
		for(int n=0; n!=CtrlNodes.length; n++) {
			CtrlNodes[n].draw(canvas, p, Window);
		}
	}
		
	// -- Touch related ------------------------------------------------------------
	
	@Override
	protected void onTouch(int x, int y, int duration, TouchType touchType) {
		if(tapCtrlNode(x, y, touchType) == true) {
			return;
		}
		
		if(tapBubbleNode(x, y, touchType) == true) {
			return;
		}
		
		if(dragBubbleNode(x, y, touchType) == true) {
			return;
		}
	}
	
	// returns true if control node was touched 
	private boolean tapCtrlNode(int x, int y, TouchType touchType) {		
		if(touchType != TouchType.TAP) {
			return false;
		}
		
		for(int n=0; n!=CtrlNodes.length; n++) {
			// remember, CtrlNodes hold the view port cooridnates 
			
			if(CtrlNodes[n].pointIsInside(x,y)) {
				// touched coordinate falls inside the node
				
				if(n==ZOOM_IN_INDEX) {
					log("Zooming in");
					Window.setZoomFactor(Window.getZoomFactor()*1.5f);
				}
				else if(n==ZOOM_OUT_INDEX) {
					log("Zooming out");
					Window.setZoomFactor(Window.getZoomFactor()*0.5f);
				}
				invalidate();
				return true;
			}
		}	
		
		// no CtrlNode was found on the tapped coordinate
		return false;
	}
	
	private boolean tapBubbleNode(int x, int y, TouchType touchType) {
		if(touchType != TouchType.TAP) {
			return false;
		}
		
		// remember, BubbleNodes hold ground plane coordinate
		int touch_x = Window.toGroundPlaneX(x);
		int touch_y = Window.toGroundPlaneY(y);
		
		int list_length = BubbleNodesToDraw.size();
		
		for(int n=0; n!=list_length; n++) {
			BubbleNode node = (BubbleNode)BubbleNodesToDraw.get(n);
			
			if(node.pointIsInside(touch_x, touch_y)) {
				// touched coordinate falls inside the node
				log(node.getName() + " is tapped");
				node.setBgColor(Color.CYAN);
				
				invalidate();
				return true;
			}
		}
		
		// no BubbleNode was found on the tapped coordinate
		return false;
	}
		
	//Node that has been pressed and now being dragged
	private BubbleNode NodeBeingHeld;
	
	private boolean dragBubbleNode(int x, int y, TouchType touchType) {
		
		if(touchType == TouchType.PRESS) {
			// remember, BubbleNodes hold ground plane coordinate
			int touch_x = Window.toGroundPlaneX(x);
			int touch_y = Window.toGroundPlaneY(y);
			
			int list_length = BubbleNodesToDraw.size();
			
			for(int n=0; n!=list_length; n++) {
				BubbleNode node = (BubbleNode)BubbleNodesToDraw.get(n);
				
				if(node.pointIsInside(touch_x, touch_y)) {
					log(node.getName() + " is pressed");
					
					Vibra.vibrate(100);
					
					// will drag this node
					NodeBeingHeld = node;
					
					// lock the view port so that the SLIDE won't move the window
					//  and is made available to this view through onTouch()
					super.lockViewport();
				}
			}
			return false;
		}
		else if(touchType == TouchType.SLIDE) {
			// SLIDE won't come in unless the lockViewport() has been called
			log(NodeBeingHeld.getName() + " is being dragged");
			
			// update the coordinate of the node being dragged
			NodeBeingHeld.setX(Window.toGroundPlaneX(x));
			NodeBeingHeld.setY(Window.toGroundPlaneY(y));
			
			invalidate();
			return true;
		}
		else if(touchType == TouchType.SLIDE_END || touchType == TouchType.FLICK) {
			// SLIDE_END / FLICK comes in at the end of SLIDE series
			log(NodeBeingHeld.getName() + " has been dragged");
			
			if(touchType == TouchType.FLICK) {
				log("flicked");
			}
			
			// slide has ended, unlock the view port
			super.unlockViewport();
			return true;
		}
		
		return false;
	}
	
	public void log(String msg) {
		Log.v("FilewayView",msg);
	}
}
