/*
  Originally written by Naoto Kaneko
  Last updated: 3.Jul.2010, buggy-and-messy version
  Latest info: http://www37.atwiki.jp/pjkyt/
 */

package com.kyt.android.view;

import java.util.*;

/**
 * This class stores DrawNodes in a matrix based on their x-y coordinate;
 *  each element corresponds to a segment of the ground plane.
 * 
 * Besides providing a storage, the point for having this class is to 
 *  offer getNodesToDraw().  This returns the list of nodes that needs to be
 *  considered when drawing the current window to a view port.  This is a critical 
 *  process for allowing to scale the FilewayView for the larger number of nodes 
 *  and larger virtual ground plane size.  
 *  
 * To find the candidates, the function only looks inside the matrix elements
 *  corresponding to the current window area and their neiboring elements.  
 *  This means a node having radius larger than the grid dimension may 
 *  encounter drawing problem.  On the other hand too large grid size 
 *  may cause rendering performance disadvantage. 
 * 
 */
public class DrawNodeList {
	// nodes are stored in the matrix based on its coordinate
	private ArrayList<DrawNode>[][] NodeList; // NodeList[row][col]
	private final int GridWidth, GridHeight;
	private final int NumOfRow, NumOfCol;
	private DrawWindow Window;
	private ArrayList<DrawNode> NodesToDraw;
	
	public DrawNodeList(DrawWindow window) {
		// just use of the screen size as a grid size
		this(window, window.getViewPortWidth(), window.getViewPortHeight());
	}
	
	public DrawNodeList(DrawWindow window, int grid_width, int grid_height) {
		Window = window;
		
		int ground = window.getGroundPlaneSize();
		GridWidth = grid_width;
		GridHeight = grid_height;

		NumOfRow = ground/GridHeight  + ((ground%GridHeight == 0) ? 0 : 1);
		NumOfCol = ground/GridWidth  + ((ground%GridWidth == 0) ? 0 : 1);
		
		NodeList = new ArrayList[NumOfRow][NumOfCol];
		
		for(int n=0; n!=NumOfRow; n++) {
			for(int m=0; m!=NumOfCol; m++) {
				NodeList[n][m] = new ArrayList<DrawNode>();
			}
		}		
		NodesToDraw = new ArrayList<DrawNode>();
	}
	
	/**
	 * Finds the column which should contain the node having specified x
	 * Returns -1 if the x doesn't fall inside the expected range
	 */
	private int findCol(int x) {
		if(x < 0 || x > Window.getGroundPlaneSize()) {
			return -1;
		}
		return (int)(x / GridWidth);
	}

	/**
	 * Finds the row which should contain the node having specified y
	 * Returns -1 if the y doesn't fall inside the expected range
	 */
	private int findRow(int y) {
		if(y < 0 || y > Window.getGroundPlaneSize()) {
			return -1;
		}
		return (int)(y / GridHeight);
	}
	
	/** 
	 * Returns false in case of such error as duplicated node
	 */
	public boolean addNode(DrawNode node) {
		if(getNode(node.getName()) != null) {
			// node with this name already exists
			return false;
		}
		int row = findRow(node.getX());
		int col = findCol(node.getY());
		
		NodeList[row][col].add(node);
		
		return true;
	}
		
	/**
	 *  Return null when no node having such name is found
	 */
	public DrawNode getNode(String name) {
		for(int n=0; n!=NumOfRow; n++) { // x
			for(int m=0; m!=NumOfCol; m++) { // y
				
				// search within this matrix element
				int list_length = NodeList[n][m].size();
				
				for(int l=0; l!=list_length; l++) {
					DrawNode node = NodeList[n][m].get(l);
					
					if(node.getName() == name) {
						return node; 
					}
				}
			}
		}
		
		return null;
	}
	
	/**
	 * Returns the number of nodes stored
	 */
	public int size() {
		int length = 0;
		
		for(int n=0; n!=NumOfRow; n++) { // x
			for(int m=0; m!=NumOfCol; m++) { // y
				length += NodeList[n][m].size();
			}
		}
		
		return length;
	}
	
	/**
	 * Returns the list of node that needs to be draw with the current window
	 */
	public ArrayList<DrawNode> getNodesToDraw() {
		// upper-left and lower-right coord of the window in the ground plane
		int win_left = Window.getX();
		int win_right = win_left + Window.getWidth();
		int win_top = Window.getY();
		int win_bottom = win_top + Window.getHeight();
		
		// find the matrix points corresponding to the current window position
		int col_from = findCol(win_left);
		int col_to = findCol(win_right);
		int row_from = findRow(win_top);
		int row_to = findRow(win_bottom);
		
		// expand the scope to neighboring elements
		if(col_from != 0) {
			col_from--;
		}
		
		if(col_to != NumOfCol-1) {
			col_to++;
		}
		
		if(row_from != 0) {
			row_from--;
		}
		
		if(row_to != NumOfRow-1) {
			row_to++;
		}

		// Make the list of nodes that are inside the window
		NodesToDraw.clear();
		
		for(int n=row_from; n!= row_to; n++) {
			for(int m=col_from; m!=col_to; m++) {
				
				// test each node inside this coordinate segment
				int list_length = NodeList[n][m].size(); 
				
				for(int l=0; l!=list_length; l++) {
					DrawNode node = NodeList[n][m].get(l);
					
					if(node.getRightBound() >= win_left && node.getLeftBound() <= win_right 
							&& node.getBottomBound() >= win_top && node.getTopBound() <= win_bottom) {
						NodesToDraw.add(node);
					}
				}
			}
		}
		
		return NodesToDraw;
	}
}
