/*
  Originally written by Naoto Kaneko
  Last updated: 3.Jul.2010, buggy-and-messy version
  Latest info: http://www37.atwiki.jp/pjkyt/
 */

package com.kyt.android.view;

import android.content.*;
import android.graphics.*;
import android.os.*;
import android.util.*;
import android.view.*;

abstract public class ViewportedView extends View {
	public enum TouchType {TAP, PRESS, SLIDE, SLIDE_END, FLICK};
	
	private DrawWindow Window;
		
	public ViewportedView(int groundPlaneSize, Context context) {
		super(context);
		Window = new DrawWindow(groundPlaneSize, context);
	}		
	
	public DrawWindow getDrawWindow() {
		return Window;
	}
	
	// -- Touch related ------------------------------------------------------------
	
	/**
	 * This function will draw position indicator bar on the bottom 
	 *  and right side of the view port
	 */
	public void drawPositionIndicator(Canvas canvas, Paint p, int barColor) {		
		p.setColor(barColor);
		p.setStrokeWidth(4);
		p.setStyle(Paint.Style.STROKE);
		
		// horizontal bar
		int start_x = (int)(Window.getViewPortWidth() * ((float)Window.getX() / Window.getGroundPlaneSize()));
		int stop_x  = start_x + (int)(Window.getViewPortWidth() * ((float)Window.getWidth() / Window.getGroundPlaneSize()));
		int start_y = Window.getViewPortHeight()-2;
		int stop_y  = start_y;
		canvas.drawLine(start_x, start_y, stop_x, stop_y, p);

		// virtical bar
		start_x = Window.getViewPortWidth()-2;
		stop_x  = start_x;
		start_y = (int)(Window.getViewPortHeight() * ((float)Window.getY() / Window.getGroundPlaneSize()));
		stop_y  = start_y + (int)(Window.getViewPortHeight() * ((float)Window.getHeight() / Window.getGroundPlaneSize()));
		canvas.drawLine(start_x, start_y, stop_x, stop_y, p);
	}

			
	// -- Touch related ------------------------------------------------------------
	
	/**
	 *  Upon TAP, PRESS, SLIDE, SLIDE_END, and FLICK, this function is invoked.
	 *  The SLIDE is only detected when lockViewPort() had been called; otherwise,
	 *   it is consumed for moving the window.  At the end of SLIDE series, when 
	 *   the finger leaves the screen, either SLIDE_END or FLICK is invoked depending
	 *   on how the finger left the screen.
	 *  The x and y are the coordinate in the view port
	 *  The duration for TAP and PRESS are the time finger had touched the screen.
	 *  The duration for MOVE is that of the most recent step in the series
	 *  The duration for SLIDE_END and FLICK are the duration of the final MOVE step
	 */
	abstract protected void onTouch(int x, int y, int duration, TouchType touchType);
	
	// When true, the slide won't move the viewport; it will instead invoke onTouch()
	private boolean ViewportLocked = false;
	
	/**
	 * By calling this function, SLIDE will not move the view port.  
	 * It will invoke onTouch() insead so that it can be used for some other purpose.
	 */
	public void lockViewport() {
		ViewportLocked = true;
	}
	
	/**
	 * Call this function to cancel the effect of lockViewPort()
	 */
	public void unlockViewport() {
		ViewportLocked = false;
	}
	
	@Override
	final public boolean onTouchEvent(MotionEvent event) {		
		int action = event.getAction();
		
		if(action == MotionEvent.ACTION_DOWN) {
			processACTION_DOWN(event);
		}
		else if(action == MotionEvent.ACTION_UP) {
			processACTION_UP(event);
		}
		else if(action == MotionEvent.ACTION_MOVE) {
			processACTION_MOVE(event);
		}
		
		return true;
	}
	
	
	private boolean PressLock = false; // flag variable to avoid multiple occurance of press during one hold
	private long TakeOffTime = 0; // time of the last MOVE; used for checking if it is a flick so that the kinetic roll should follow
	private KineticRoll Kr = null; // AsyncTask to implement the kinetic roll
	
	private void processACTION_DOWN(MotionEvent event) {
		// see how they are updated when MotionEvent.MOVE occurs
		TakeOffTime = event.getEventTime(); 
		PressLock = false;
		
		if(Kr != null) {
			// kinetic roll is on-going; stop it
			Kr.cancel(true);
			Kr = null;
		}
	}
	
	private final int TAP_THRESHOLD = 200; // it is a TAP if it is less than this; in ms
	private final int FLICK_THRESHOLD = 50; // it is a FLICK when the finger leaves screen faster than this after a slide; in ms
	
	private void processACTION_UP(MotionEvent event) {
		// accumulated duration since the finger has touched down on the screen
		int duration = (int)(event.getEventTime()-event.getDownTime());
		
		int x = (int)event.getX();
		int y = (int)event.getY();
		
		int take_off_speed = (int)(event.getEventTime() - TakeOffTime);
		
		log("UP duration = " + duration + "ms ("+x+","+y+")");
		log("Take off speed = " + take_off_speed + "ms");
		
		if(duration == take_off_speed && duration < TAP_THRESHOLD) {
			// TAP detected
			// "duration == take_off_speed" means MotionEvent.MOVE had never occured
			onTouch(x,y,duration,TouchType.TAP);
		}
		else if(take_off_speed < FLICK_THRESHOLD) {
			// FLICK detected
			
			if(ViewportLocked == true) {
				// user had called lockViewport()
				onTouch(x,y,StepDuration,TouchType.FLICK);
			}
			else {
				Kr = new KineticRoll();
				Kr.execute(new Integer(XDisp),new Integer(YDisp),new Integer(StepDuration));
			}
		}
		else if(ViewportLocked == true) {
			// user had called lockViewport()
			// SLIDE_END detected
			onTouch(x,y,StepDuration,TouchType.SLIDE_END);
		}
	}
	
	private final int PRESS_THRESHOLD = 1000; //it is a PRESS when the tap holds for this; in ms
	private final float SLIDE_BOOST_FACTOR = 1.3f; //ugly parameter to make the touch feel right
	private int XDisp, YDisp; // Displacement of the most recent MOVE step; in viewport metric
	private int StepDuration; // Duration of the most recent MOVE step in ms
	
	private void processACTION_MOVE(MotionEvent event) {
		// accumulated duration since the finger has touched down on the screen
		int duration = (int)(event.getEventTime()-event.getDownTime());
		
		int x = (int)event.getX();
		int y = (int)event.getY();
		
		//log("MOVE duration = " + duration + "ms ("+x+","+y+")");
		
		int history_size = event.getHistorySize();
		
		if(history_size == 0) {
			if(duration > PRESS_THRESHOLD && PressLock == false) {
				// PRESS detected
				PressLock = true;
				onTouch(x,y,duration,TouchType.PRESS); // check if there is any node here
			}
		}
		else { 
			// SLIDE detected
			StepDuration = (int)(event.getEventTime() - TakeOffTime);
			TakeOffTime = event.getEventTime();
			
			log("SLIDE ("+x+","+y+")");
			//
			//for(int n=0; n!=history_size; n++) {
			//	log(n + "("+event.getHistoricalX(n)+","+event.getHistoricalY(n)+")");
			//}
			
			// calculate the displacement in the view port metric
			XDisp = (int)((x - event.getHistoricalX(0))* SLIDE_BOOST_FACTOR);
			YDisp = (int)((y - event.getHistoricalY(0))* SLIDE_BOOST_FACTOR);
			
			log("x_disp="+XDisp+" y_disp="+YDisp+" step-duration="+StepDuration);
			
			if(ViewportLocked == false) {
				// move the window
				int x_new = Window.getX()-Window.toGroundPlaneMetric(XDisp);
				int y_new = Window.getY()-Window.toGroundPlaneMetric(YDisp);
				
				Window.setXY(x_new, y_new); 
				invalidate(); // render with updated window
			}
			else {
				// user had called lockViewport()
				onTouch(x, y, duration, TouchType.SLIDE);
			}
		}
	}
	
	private class KineticRoll extends AsyncTask<Integer,Void,Void> {
		private boolean Canceled = false;
		private final float FLUCTUATION_RATE = 1.01f;
		
		// rollInfo[0]: last x displacement
		// rollInfo[1]: last y displacement
		// rollInfo[2]: span between moves in ms
		@Override
		protected Void doInBackground(Integer... rollInfo) {
			// convert the displacement to the ground plane metric
			int x_disp = Window.toGroundPlaneMetric(rollInfo[0]);
			int y_disp = Window.toGroundPlaneMetric(rollInfo[1]);
			int duration = rollInfo[2];
			
			while(true) {
				Window.setXY(Window.getX()-x_disp, Window.getY()-y_disp);
				publishProgress((Void[])null);
				
				x_disp /= FLUCTUATION_RATE;
				y_disp /= FLUCTUATION_RATE;
				
				if(x_disp == 0 && y_disp == 0) {
					break;
				}
				
				if(Canceled == true) {
					break;
				}
				
				try {
					Thread.sleep(duration);
				} catch (InterruptedException e) {
					
				}
			}
			return null;
		}

		@Override
		protected void onPostExecute(Void result) { }
		
		@Override
		protected void onProgressUpdate(Void... progress) {
			invalidate();
		}

		@Override
		protected void onCancelled() {
			Canceled = true;
		}
	}
	
	//-- Log related ------------------------------------------------------------
	
	private boolean ShowLog = false;
	
	public void showLog() {
		ShowLog = true;
	}
	
	private void log(String msg) {
		if(ShowLog == true) {
			Log.v("ViewPortedView",msg);
		}
	}
}
